from .xlrd_gsheet_reader import (
    download_sheet_from_google_docs,
    parse_xls_workbook,
)
from .data_utils import to_dicts


class FastGSheet:
    def __init__(self, sheet_key, auth):
        content = download_sheet_from_google_docs(sheet_key, auth)
        self.data = {
            key: FastSheet(value) for key, value in parse_xls_workbook(content).items()
        }

    def worksheet(self, name):
        return self.data[name]


class FastSheet:
    def __init__(self, data):
        self.data = data

    def get_all_records(self, numericise_ignore=None):
        return to_dicts(self.data)

    def get_all_values(self, numericise_ignore=None):
        return self.data



from sparkles import to_dicts


def test_to_dicts_returns_empty_list_when_empty():
    assert to_dicts(None) == []


def test_to_dicts_returns_empty_list_when_only_one_row():
    assert to_dicts([["a", "b"]]) == []


def test_to_dicts_returns_one_row_when_header_and_data_present():
    assert to_dicts([["a", "b"], [1, 2]]) == [{"a": 1, "b": 2}]


def test_to_dicts_handles_iterable_instead_of_list():
    gen = (x for x in [["a", "b"], [1, 2]])

    assert to_dicts(gen) == [{"a": 1, "b": 2}]



"""Just utilities made by Jim Vogel"""
import itertools
from collections import namedtuple
from datetime import datetime
from functools import wraps
from typing import Iterable, List
from subprocess import check_output

from loguru import logger
import pytz
import requests

from .mongo_tools import clean_json
from .data_utils import to_dicts

__version__ = "0.1.20"


def logged_user(logger=logger):
    def logging_user_wrapper(func):
        function_name = func.__name__

        @wraps(func)
        def log(*args, **kwargs):
            user = kwargs.get("current_user")
            username = user.username if user else "Unknown User"

            logger.info(f"{username} ran {function_name}")
            return func(*args, **kwargs)

        return log

    return logging_user_wrapper


def parse_float(value: str) -> float:
    value = value.strip()
    if value == "-":
        return 0.0

    return float(value.replace(",", "").replace("$", ""))


def parse_int(value: str):
    float_value = parse_float(value)
    if not float_value.is_integer():
        raise ValueError(f"{value} is not an integer")
    return int(float_value)


def now_utc():
    return datetime.now(pytz.utc).replace(second=0, microsecond=0, tzinfo=None)


def trunc_date(dt):
    return dt.replace(hour=0, minute=0, second=0, microsecond=0)


def chunked(iterable, size):
    it = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(it, size))
        if not chunk:
            break
        yield chunk


def add_key_from_id(data: Iterable[dict]):
    for row in data:
        row["key"] = row["id"]


def quick_search(query, model, fields=None, base_raw_query=None):
    fields = fields or model.quick_search_fields
    query = query.lower()
    ret = []
    for title, field_name in fields:
        values_with_query_in_it = [
            f
            for f in model.objects(__raw__=base_raw_query or {}).distinct(
                field=field_name
            )
            if query in f.lower()
        ]

        options = []
        for value in values_with_query_in_it:
            row_count = len(
                model.objects(__raw__={**(base_raw_query or {}), field_name: value})
            )
            options.append(
                {"label": value, "count": row_count, "filter": {field_name: value}}
            )
        if options:
            ret.append({"title": title, "options": options})

    return ret


GpsCoords = namedtuple("GpsCoords", "lat,lon")


def build_send_slack_message(prefix, slack_webhook):
    def send_slack_message(message):
        from notifiers import get_notifier

        message = f"{prefix} | {message}"

        lines = message.splitlines(keepends=True)
        line_count = len(lines)
        if line_count > 10:
            url, worked = send_to_haste_bin(message)
            if worked:
                slack_message = "".join(
                    [lines[0], f"{line_count} lines long, pasted here: {url}"]
                )
            else:
                slack_message = message + "\n" + url
        else:
            slack_message = message

        notifier = get_notifier("slack")
        try:
            notifier.notify(webhook_url=slack_webhook, message=slack_message)
        except Exception:
            pass

    return send_slack_message


HASTEBIN = "https://hastebin.com/"


def send_to_haste_bin(message):
    try:
        response = requests.post(f"{HASTEBIN}documents", message.encode("utf-8"))
        return f'{HASTEBIN}{response.json()["key"]}', True
    except Exception:
        import traceback

        return f"Failed to send to hastebin, {traceback.format_exc()}", False


def parse_csl(value, lower=True):
    """
    reads a string that is a comma separated list and converts it into a list of strings
    stripping out white space along the way
    :param value: the string to parse
    :param lower: should we also lowercase everything (default true)
    :return: a list of strings
    """
    value = value.strip()
    if not value:
        return []
    if lower:
        return [t.strip() for t in value.lower().strip().split(",")]
    return [t.strip() for t in value.strip().split(",")]


def sorted_groupby(iterable, key, value=lambda x: x):
    """
    just like regular pythons regular group by except it sorts for you
    :param iterable: the thing to iterate through
    :param key: the key to both sort and group by
    :param value: a function to apply to the group before returning (defaults to groupby iterator)
    :return: iterable of tuples if (key_results, grouped_values)
    """
    return (
        (k, value(v))
        for k, v in itertools.groupby(sorted(iter(iterable), key=key), key=key)
    )


def counter(start=0, step=1):
    count = start
    while True:
        yield count
        count += step


def pairwise(iterable: Iterable):
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def get_stdout_lines(command: List[str]) -> List[str]:
    return check_output(command).decode().splitlines()



"""Provide a custom uvicorn worker that can be configured."""
from starlette.config import Config
from uvicorn.workers import UvicornWorker


config = Config()


class ConfigurableWorker(UvicornWorker):
    #: dict: Set the equivalent of uvicorn command line options as keys.
    CONFIG_KWARGS = {
        "root_path": config("SCRIPT_NAME", default=""),
        "proxy_headers": True,
    }



from typing import Iterable


def split_header_and_data(rows: Iterable[list]):
    row_iter = iter(rows)
    header = next(row_iter)
    return header, row_iter


def list_of_dicts(header, data):
    return [dict(zip(header, row)) for row in data]


def to_dicts(ds):
    if not ds:
        return []
    header, data = split_header_and_data(ds)
    return list_of_dicts(header, data)



import pytest
from sparkles import logged_user


class MockLogger:
    last_line = None

    def info(self, message):
        self.last_line = message


@pytest.fixture
def logger():
    return MockLogger()


@pytest.fixture
def local_logged_user(logger):
    return logged_user(logger=logger)


@pytest.fixture
def logged_do_nothing(local_logged_user):
    @local_logged_user
    def do_nothing(*args, **kwargs):
        pass

    return do_nothing


class MockUser:
    username = "mock"


@pytest.fixture
def mock_user():
    return MockUser()


def test_logged_user_returns_a_function(local_logged_user):
    assert callable(local_logged_user)


def test_logged_user_doesnt_fail_when_no_user(logged_do_nothing):
    logged_do_nothing()


def test_logged_user_logs_a_line_when_method_called(logger, logged_do_nothing):
    assert logger.last_line is None
    logged_do_nothing()
    assert logger.last_line


def test_logged_user_logs_include_username(mock_user, logger, logged_do_nothing):
    logged_do_nothing(current_user=mock_user)
    assert "mock" in logger.last_line



from typing import Dict, Any


def replace__id_with_id(r: Dict[str, Any]):
    if "_id" in r:
        r["id"] = str(r["_id"])
        del r["_id"]


def mongo_query_to_list_of_dicts(object_query):
    return [o.to_mongo().to_dict() for o in object_query]


def replace_all_ids(query_results):
    for result in query_results:
        replace__id_with_id(result)


def clean_json(mongoengine_query):
    query_results = mongo_query_to_list_of_dicts(mongoengine_query)
    replace_all_ids(query_results)
    return query_results



#!/usr/bin/env python3
from sparkles import get_stdout_lines

import os
import sys


def need_to_commit():
    status_lines = get_stdout_lines(["git", "status"])
    return "nothing to commit, working tree clean" not in status_lines


def get_commit_message():
    args_contain_commit_message = len(sys.argv) > 1

    commit_message = (
        get_commit_message_from_arguments()
        if args_contain_commit_message
        else get_commit_message_interactive()
    )

    return escape_quotes_for_bash(commit_message)


def get_commit_message_interactive():
    return input("Commit message? ")


def get_commit_message_from_arguments():
    return " ".join(sys.argv[1:])


def escape_quotes_for_bash(bash_line):
    return bash_line.replace("'", "'\\''")


def commit_changes():
    os.system("git add .")
    os.system(f'git commit -m "{get_commit_message()}"')


def bump_version():
    os.system("bump2version patch")
    os.system("git push")


def publish():
    os.system("flit publish")


if __name__ == "__main__":
    if need_to_commit():
        commit_changes()
    bump_version()
    publish()



from pprint import pprint
from time import sleep

from sparkles.google_sheet import get_book


def get_worksheet_names(book):
    return {sheet.title for sheet in book}


def foo(sheet_names, master, target):
    pass


def run(master_sheet_key, target_sheet_key):
    master_book = get_book(master_sheet_key)
    target_book = get_book(target_sheet_key)

    master_sheet_names = get_worksheet_names(master_book)
    target_sheet_names = get_worksheet_names(target_book)

    missing_sheets = master_sheet_names - target_sheet_names

    ret = {}
    for name in master_sheet_names.intersection(target_sheet_names):
        print(f"Doing {name}")
        master_ws = master_book.worksheet(name)
        target_ws = target_book.worksheet(name)
        column_diff = set(master_ws.row_values(1)) - set(target_ws.row_values(1))
        print(column_diff)
        ret[name] = column_diff
        sleep(10)

    return {"missing_sheets": missing_sheets, "missing_columns": ret}


if __name__ == "__main__":
    pprint(
        run(
            "1IBGy2cY5q79en7OEs1b0QoAZlTaLrw6KoAF-O75XnSE",
            "16L2EQ4zA9d2FrEk7vdt3xiVclZFr1mN-lOFSv6lmSBQ",
        )
    )



from io import BytesIO

import xlrd
import requests
from .google_sheet import auth


def google_url_from_sheet_key(sheet_key):
    return (
        f"https://www.googleapis.com/drive/v3/files/{sheet_key}/export?"
        f"mimeType=application%2Fvnd.openxmlformats-officedocument.spreadsheetml.sheet"
    )


def download_sheet_from_google_docs(sheet_key, auth):
    access_token = get_gspread_access_token(auth)
    url = google_url_from_sheet_key(sheet_key)
    res = requests.get(url, headers={"Authorization": f"Bearer {access_token}"})
    return res.content


def get_gspread_access_token(auth):
    client = auth or auth()
    if not client.auth.token:
        client.login()
    return client.auth.token


def parse_xls_workbook(content):
    workbook = xlrd.open_workbook(file_contents=BytesIO(content).read())
    return {sheet.name: list(data_rows(sheet)) for sheet in workbook.sheets()}


def data_rows(sheet):
    for row in sheet.get_rows():
        if row[0].ctype == xlrd.XL_CELL_EMPTY:
            break
        yield [parse_excel_cell(column) for column in row]


def parse_excel_date(value):
    return xlrd.xldate.xldate_as_datetime(value, 0).isoformat()


parsers = {
    xlrd.XL_CELL_BOOLEAN: lambda x: "True" if x else "False",
    xlrd.XL_CELL_DATE: parse_excel_date,
    xlrd.XL_CELL_NUMBER: lambda v: str(int(v) if v.is_integer() else v),
    xlrd.XL_CELL_TEXT: lambda v: v,
    xlrd.XL_CELL_EMPTY: lambda v: v,
    xlrd.XL_CELL_BLANK: lambda v: v,
    xlrd.XL_CELL_ERROR: lambda v: "",
}


def parse_excel_cell(column: xlrd.sheet.Cell):
    parser = parsers[column.ctype]
    return parser(column.value)



import pytest
from sparkles import counter


@pytest.fixture
def a_counter():
    return counter(12, 6)


def test_first_counter_is_start():
    q = next(counter(12))
    assert q == 12


def test_second_counter_is_start_plus_step():
    start = 5
    step = 12
    expected = start + step
    c = counter(start, step)
    next(c)
    q = next(c)
    assert q == expected



import os


def find_file_above(file_to_find: str, path: str = None):
    starting_directory = path or os.getcwd()

    tree = FileTree(starting_directory)

    while not tree.is_at_filesystem_root:
        if file_to_find in tree.current_files:
            return f"{tree.current_directory}/{file_to_find}"
        tree.move_up_one_level()


class FileTree:
    def __init__(self, starting_directory):
        self.current_directory = starting_directory

    @property
    def current_files(self):
        return os.listdir(self.current_directory)

    @property
    def parent_dir(self):
        return os.path.dirname(self.current_directory)

    @property
    def is_at_filesystem_root(self):
        return self.current_directory == self.parent_dir

    def move_up_one_level(self):
        self.current_directory = self.parent_dir



import pytest

from sparkles import sorted_groupby


def test_takes_an_iterable_and_a_key():
    sorted_groupby(iterable="abcd", key=lambda x: x)


def test_throws_exception_if_iterable_is_not_iterable():
    with pytest.raises(TypeError):
        sorted_groupby(iterable=1, key=lambda x: x)


def test_groups_unsorted():
    assert dict(sorted_groupby(iterable="aba", key=lambda x: x, value=list)) == {
        "a": ["a", "a"],
        "b": ["b"],
    }



import logging
import os
from pathlib import Path

from gspread import WorksheetNotFound, authorize
from loguru import logger
from oauth2client.service_account import ServiceAccountCredentials

from .file_utils import find_file_above


class Config:
    google_credentials_file = "service_account.json"
    default_google_sheet_key = ""
    ignored_sheets = ["Results", "Status"]


config = Config()


def ignored_sheets():
    return [tab_name.casefold() for tab_name in config.ignored_sheets]


def auth(credential_file=None):
    credential_file = credential_file or config.google_credentials_file
    path = absolute_path_for(credential_file)
    logger.info(f"Authorizing google with credential file {path}")
    credentials = load_gspread_credentials(path)
    return authorize(credentials)


def load_gspread_credentials(absolute_path):
    scope = [
        "https://spreadsheets.google.com/feeds",
        "https://www.googleapis.com/auth/drive",
    ]
    return ServiceAccountCredentials.from_json_keyfile_name(absolute_path, scope)


def absolute_path_for(original_credential_file):
    if os.sep in original_credential_file:
        credential_file = Path(__file__).parent.parent / original_credential_file
    else:
        credential_file = find_file_above(original_credential_file, os.getcwd())

    if not credential_file:
        raise Exception(
            f"Could not find a credential file {original_credential_file} above {os.getcwd()}"
        )
    return credential_file


def get_book(sheet_key, gc=None):
    if gc is None:
        gc = auth()
    elif type(gc) == str:
        gc = auth(gc)
    logging.info(f"Opening Google Sheet {sheet_key}")
    book = gc.open_by_key(sheet_key)
    logging.info(f"Google Sheet Title is {book.title}")
    return book


def get_fast_book(sheet_key, gc=None):
    if gc is None:
        gc = auth()
    elif type(gc) == str:
        gc = auth(gc)
    logging.info(f"Opening Google Sheet {sheet_key}")
    from .fast_gsheet import FastGSheet

    book = FastGSheet(sheet_key, gc)
    return book


def data_sheets(book):
    """Lists all worksheets that aren't in ignored sheets"""
    for worksheet in book:
        title = worksheet.title
        if title.casefold() in ignored_sheets():
            continue
        yield worksheet


def clean_worksheet_data(worksheet, log):
    log(f"Loading sheet {worksheet.title}")
    data = worksheet.get_all_values()
    data[0] = [str(col).strip() for col in data[0]]  # strip whitespace on column names
    return data


def to_dataset(book, log=lambda x: None):
    log("Loading sheets")
    return {
        worksheet.title: clean_worksheet_data(worksheet, log)
        for worksheet in data_sheets(book)
    }


def data_set(sheet_key, log=lambda x: None):
    book = get_book(sheet_key)
    return book.title, to_dataset(book, log)


def get_sheet(book, name):

    try:
        return book.worksheet(name)
    except WorksheetNotFound:
        logger.exception(
            f"Tried to load a sheet named: {name} but there is not sheet named that here: {book.url}"
        )
        return


def get_sheet_data(name, book=None):
    book = book or get_book(config.default_google_sheet_key)
    if sheet := get_sheet(book, name):
        return sheet.get_all_records()
    return []



import pytest
from sparkles import parse_float, parse_int

valid_float_inputs = [
    ("3,000.00", 3000.0),
    ("-", 0.0),
    (" -", 0.0),
    ("- ", 0.0),
    (" 3,000.00", 3000.0),
    (" 3,000.00 ", 3000.0),
    (" $3,000.00 ", 3000.0),
]


@pytest.mark.parametrize("float_str_value,expected", valid_float_inputs)
def test_parsing_valid_floats(float_str_value, expected):
    assert parse_float(float_str_value) == expected


invalid_float_inputs = [
    "A3,000.00",
    " MIKE ",
]


@pytest.mark.parametrize("float_str_value", invalid_float_inputs)
def test_parsing_valid_floats(float_str_value):
    with pytest.raises(ValueError):
        parse_float(float_str_value)


valid_int_inputs = [
    ("3,000.00", 3000),
    ("-", 0),
    (" -", 0),
    ("- ", 0),
    (" 3,000.00", 3000),
    (" 3,000.00 ", 3000),
    (" $3,000.00 ", 3000),
]


@pytest.mark.parametrize("int_str_value,expected", valid_int_inputs)
def test_parsing_valid_ints(int_str_value, expected):
    int_value = parse_int(int_str_value)
    assert int_value == expected
    assert isinstance(int_value, int), "int_value response is not an int"


invalid_int_inputs = [
    "A3,000.50",
    " MIKE ",
    "3000.5",
]


@pytest.mark.parametrize("int_str_value", invalid_int_inputs)
def test_parsing_valid_floats(int_str_value):
    with pytest.raises(ValueError):
        parse_int(int_str_value)



from sparkles import parse_csl


def test_single_item():
    assert parse_csl("test") == ["test"]


def test_no_items():
    assert parse_csl("") == []


def test_2_items():
    assert parse_csl("bob,villa") == ["bob", "villa"]


def test_lowercase():
    assert parse_csl("Bob") == ["bob"]


def test_no_lowercase():
    assert parse_csl("Bob", lower=False) == ["Bob"]



import pytest
from sparkles.mongo_tools import (
    replace__id_with_id,
    mongo_query_to_list_of_dicts,
    clean_json,
)


class MockMongoEngObj:
    def to_mongo(self):
        return MockInner()


class MockInner:
    def to_dict(self):
        return {"_id": 123}


@pytest.fixture
def mongoengine_query():
    return [MockMongoEngObj()]


def test_replace__id_with_id_actually_replaces():
    data = {"_id": "3"}
    replace__id_with_id(data)
    assert data == {"id": "3"}


def test_replace__id_with_id_doesnt_fail_without_an_id():
    data = {"qid": 3}
    replace__id_with_id(data)
    assert data == {"qid": 3}


def test_replace__id_with_id_stringifies():
    data = {"_id": 3}
    replace__id_with_id(data)
    assert data == {"id": "3"}


def test_converting_mongo_query_result_is_list(mongoengine_query):
    result = mongo_query_to_list_of_dicts(mongoengine_query)
    assert type(result) is list


def test_converting_mongo_query_first_result_is_dict(mongoengine_query):
    result = mongo_query_to_list_of_dicts(mongoengine_query)
    assert type(result[0]) is dict


@pytest.fixture
def cleaned_mongo_json(mongoengine_query):
    return clean_json(mongoengine_query)


def test_clean_json_result_is_a_list(cleaned_mongo_json):
    assert type(cleaned_mongo_json) is list


def test_clean_json_first_result_is_a_dict(cleaned_mongo_json):
    assert type(cleaned_mongo_json[0]) is dict


def test_clean_json_results_dont_have__id(cleaned_mongo_json):
    assert not any("_id" in r for r in cleaned_mongo_json)
